/**
 * Alipay.com Inc.
 * Copyright (c) 2004-2017 All Rights Reserved.
 */
package com.alipay.fc.cschat.web.openapi.controller.test;

import com.alibaba.fastjson.JSONObject;
import com.alipay.fc.cschat.common.util.logger.LoggerUtil;
import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.HttpConnectionManager;
import org.apache.commons.httpclient.HttpException;
import org.apache.commons.httpclient.HttpStatus;
import org.apache.commons.httpclient.MultiThreadedHttpConnectionManager;
import org.apache.commons.httpclient.methods.GetMethod;
import org.apache.commons.httpclient.methods.PostMethod;
import org.apache.commons.httpclient.methods.RequestEntity;
import org.apache.commons.httpclient.methods.StringRequestEntity;
import org.apache.commons.httpclient.methods.multipart.FilePart;
import org.apache.commons.httpclient.methods.multipart.MultipartRequestEntity;
import org.apache.commons.httpclient.methods.multipart.Part;
import org.apache.commons.httpclient.methods.multipart.StringPart;
import org.apache.commons.httpclient.params.HttpConnectionManagerParams;
import org.apache.commons.lang.ArrayUtils;
import org.apache.log4j.Logger;
import org.springframework.web.multipart.MultipartFile;

import javax.crypto.Mac;
import javax.crypto.spec.SecretKeySpec;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;

/**
 * @author fushan.cfs
 * @version $Id: CustomerService.java, v 0.1 2017年08月23日 下午4:23 alps Exp $
 */
public class CustomerService {

    private static final Logger logger       = Logger
            .getLogger(CustomerService.class);
    /** http请求的header设置 */
    private static final String CONTENT_TYPE = "application/json";
    private static final String CHARSET      = "UTF-8";

    /** 用于发送消息的httpClient */
    private HttpClient httpClient;
    /**
     * httpClient的配置内容
     *
     */
    private int maxConnPerHost = 40;
    private int maxTotalConn = 200;
    /**
     * 默认等待连接建立超时，单位:毫秒
     */
    private int connectionTimeout = 1000;
    /**
     * 默认等待数据返回超时，单位:毫秒
     */
    private int soTimeout = 10000;
    /**
     * 默认请求连接池连接超时,单位:毫秒
     */
    private int connectionManagerTimeout = 1000;
    /**
     * 请求的数据内容
     */
    private RequestInfo requestInfo;



    /**
     * 用于初始化服务的构造函数
     *
     * 参数内容全部需要从云客服这里获取
     * 如果不知道参数具体应该填写什么内容
     * 请参考文档或联系云客服
     *
     * @param tntInstId 租户id
     * @param scene 聊天窗编码（涉及到技能组)
     * @param key 私钥用于加密 (谨慎保存不要放到消息体内传送给云客服)
     */
    public CustomerService(String tntInstId, String scene, byte[] key, String url) {
        this.requestInfo = new RequestInfo();
        requestInfo.setTntInstId(tntInstId);
        requestInfo.setScene(scene);
        requestInfo.setKey(key);
        requestInfo.setUrl(url);
        createHttpClient();
    }

    /**
     * 构造httpClient用于对消息进行发送
     */
    private void createHttpClient(){
        HttpConnectionManager httpConnectionManager = new MultiThreadedHttpConnectionManager();
        HttpConnectionManagerParams connectionManagerParams = new HttpConnectionManagerParams();
        connectionManagerParams.setConnectionTimeout(connectionTimeout);
        connectionManagerParams.setSoTimeout(soTimeout);
        connectionManagerParams.setMaxTotalConnections(maxTotalConn);
        connectionManagerParams.setDefaultMaxConnectionsPerHost(maxConnPerHost);
        httpConnectionManager.setParams(connectionManagerParams);
        this.httpClient = new HttpClient(httpConnectionManager);
        this.httpClient.getParams().setConnectionManagerTimeout(connectionManagerTimeout);
    }


    /**
     * hmac_sha1算法计算内容摘要
     *
     * value的生成流程请看详细文档对于不同的消息类型构造方式不同
     * <ul>
     *   <li>文本消息：toJSONString( MessageContent ) + timestamp</li>
     *   <li>base64文件：base64字符串 + timestamp</li>
     * </ul>
     *
     * @param value 原始数据的字符串
     * @param keyBytes   密钥
     * @return 加密后的字符串
     */
    public String hmac_sha1(String value, byte[] keyBytes) {
        try {
            SecretKeySpec signingKey = new SecretKeySpec(keyBytes, "HmacSHA1");
            Mac mac = Mac.getInstance("HmacSHA1");
            mac.init(signingKey);
            byte[] rawHmac = mac.doFinal(value.getBytes("utf-8"));
            String digest = toHexString(rawHmac);
            return digest;
        } catch (Exception ex) {
            return null;
        }
    }

    /**
     * hmac_sha1算法计算内容摘要
     *
     * 主要用于二进制的文件摘要计算
     * <ul>
     *     <li>二进制文件消息： byte[] file + byte timestamp </li>
     * </ul>
     * @param bytes 需要加密的二进制数组
     * @param keyBytes 私钥
     * @return 加密后的字符串
     */
    private String hmac_sha1(byte[] bytes, byte[] keyBytes) {
        try {
            SecretKeySpec signingKey = new SecretKeySpec(keyBytes, "HmacSHA1");
            Mac mac = Mac.getInstance("HmacSHA1");
            mac.init(signingKey);
            byte[] rawHmac = mac.doFinal(bytes);
            String digest = toHexString(rawHmac);
            return digest;
        } catch (Exception ex) {
            return null;
        }
    }


    /**
     * 二进制字节数组转十六进制
     *
     * @param v 二进制数组内容
     * @return 十六进制字符串
     */
    public  String toHexString(byte[] v) {
        StringBuilder sb = new StringBuilder(v.length * 2);
        for (int i = 0; i < v.length; i++) {
            int b = v[i] & 0xFF;
            sb.append(CustomerServiceConfig.HEX_DIGITS.charAt(b >>> 4)).append(
                CustomerServiceConfig.HEX_DIGITS.charAt(b & 0xF));
        }
        return sb.toString();
    }

    /**
     * 构造文本消息的请求详细内容
     *
     * @param messageContent 发送消息body
     */
    private void constructRequestInfo(JSONObject messageContent) {
        //初始化时间戳 由于后台会判断时间戳有效期 请一定使用当前时间戳
        long timestamp = System.currentTimeMillis();
        //序列化消息体
        String requestContentStr = JSONObject.toJSONString(messageContent);
        //生成签名
        String digest = hmac_sha1(requestContentStr + timestamp, requestInfo.getKey());
        requestInfo.setDigest(digest);
        requestInfo.setRequestContent(requestContentStr);
        requestInfo.setTimestamp(timestamp);
        requestInfo.constructUrl(CustomerServiceConfig.FORWARD_MESSAGE_API_PATH);
    }


    /**
     * 构造普通文件上传的消息体详细内容
     * @param fileContent 文件详情
     * @throws IOException
     */
    private void constructFileRequestInfo(FileContent fileContent) throws IOException {
        MultipartFile file = fileContent.getFile();

        long timestamp = System.currentTimeMillis();
        byte[] fileBytes = file.getBytes();
        byte[] timeBytes = Long.toString(timestamp).getBytes();
        byte[] digestBytes = ArrayUtils.addAll(fileBytes, timeBytes);
        String digest = hmac_sha1(digestBytes, requestInfo.getKey());
        requestInfo.setDigest(digest);
        requestInfo.setTimestamp(timestamp);
        requestInfo.constructUrl(CustomerServiceConfig.UPLOAD_FILE_API_PATH);
    }

    /**
     * 构造base64文件上传的消息详细内容
     * @param fileContent 文件详情
     */
    private void constructBase64RequestInfo(FileContent fileContent) {
        String base64File = fileContent.getBase64File();
        long timestamp = System.currentTimeMillis();

        byte[] fileBytes = base64File.getBytes();
        byte[] timeBytes = String.valueOf(timestamp).getBytes();
        byte[] digestBytes = ArrayUtils.addAll(fileBytes, timeBytes);

        String digest = hmac_sha1(digestBytes, requestInfo.getKey());
        requestInfo.setDigest(digest);
        requestInfo.setTimestamp(timestamp);
        requestInfo.constructUrl(CustomerServiceConfig.UPLOAD_FILE_API_PATH);
    }

    /**
     * 上传文件接口 该接口上传file类型文件
     * base64编码过后会变大，请注意文件上传大小限制
     *
     * @param fileContent 文件上传的内容
     * @return 上传结果
     * @throws Exception
     */
    public String uploadMedia(FileContent fileContent) throws Exception {
        // 判断文件类型是否有效
        String fileType = fileContent.getType();
        if (FileTypeEnum.getByFileType(fileType) == null) {
            throw new EnumConstantNotPresentException(FileTypeEnum.class, "[ Do not support the file type ]");
        }
        //检查是否有必须填写的字段为空
        if (!fileContent.valid()){
            throw new RuntimeException("file content valid check error");
        }
        constructFileRequestInfo(fileContent);
        PostMethod postMethod = new PostMethod(requestInfo.getUrl());

        //构造POST body内容
        byte[] multiFileBytes = fileContent.getFile().getBytes();
        File file = null;
        try {
            file = File.createTempFile(fileContent.getFileName(), fileContent.getType());
            FileOutputStream fo = new FileOutputStream(file);
            fo.write(multiFileBytes);
            fo.close();
        } catch (IOException e) {
            LoggerUtil.warn(logger, "create tmp file failed!", e);
            return null;
        }
        FilePart filePart = new FilePart("file", file);

        StringPart typePart = new StringPart("type", fileContent.getType(), CHARSET);
        StringPart fileNamePart = new StringPart("fileName", fileContent.getFileName(), CHARSET);
        StringPart timestampPart = new StringPart("timestamp", Long.toString(requestInfo.getTimestamp()), CHARSET);
        Part[] parts = { typePart, fileNamePart, timestampPart, filePart };

        MultipartRequestEntity multipartRequestEntity = new MultipartRequestEntity(parts, postMethod.getParams());
        postMethod.setRequestEntity(multipartRequestEntity);
        try {
            this.httpClient.executeMethod(postMethod);
            String responseStr = postMethod.getResponseBodyAsString();
            return responseStr;
        } catch (HttpException he) {
            throw new HttpException("[ send message error ]", he);
        } catch (IOException ie) {
            throw new IOException("[ send message or get response error ]", ie);
        } finally {
            postMethod.releaseConnection();
        }
    }

    /**
     * 上传base64格式的文件
     *
     * @param fileContent
     * @return
     * @throws IOException
     */
    public String uploadBase64Media(FileContent fileContent) throws IOException {
        String fileType = fileContent.getType();
        if (FileTypeEnum.getByFileType(fileType) == null) {
            throw new EnumConstantNotPresentException(FileTypeEnum.class, "[ Do not support the file type ]");
        }
        //检查是否有必须填写的字段为空
        if (!fileContent.valid()){
            return null;
        }
        constructBase64RequestInfo(fileContent);
        PostMethod postMethod = new PostMethod(requestInfo.getUrl());
        StringPart base64FilePart = new StringPart("base64File", fileContent.getBase64File(), CHARSET);
        StringPart typePart = new StringPart("type", fileContent.getType(), CHARSET);
        StringPart fileNamePart = new StringPart("fileName", fileContent.getFileName(), CHARSET);
        StringPart timestampPart = new StringPart("timestamp", Long.toString(requestInfo.getTimestamp()), CHARSET);
        Part[] parts = {typePart, fileNamePart, timestampPart, base64FilePart};
        MultipartRequestEntity multipartRequestEntity = new MultipartRequestEntity(parts, postMethod.getParams());
        try {
            postMethod.setRequestEntity(multipartRequestEntity);
            int status = this.httpClient.executeMethod(postMethod);
            if (status != HttpStatus.SC_OK) {
                return postMethod.getResponseBodyAsString();
            }
            //这里是不是在response body有内容需要看一下
            return postMethod.getResponseBodyAsString();
        }catch (UnsupportedEncodingException e) {
            throw new UnsupportedOperationException("unSupport encoding : " + CHARSET, e);
        }catch (HttpException he) {
            throw new HttpException("[ send message error ]", he);
        } catch (IOException ie) {
            throw new IOException("[send message or get response error ]", ie);
        } finally {
            postMethod.releaseConnection();
        }
    }


    /**
     * 发送信息到在线客服Server
     *
     * @param MessageContent 消息的Body内容
     * @return 返回发送信息结果
     * @throws Exception
     */
    public JSONObject forwardMessage(JSONObject MessageContent) throws Exception {
        constructRequestInfo(MessageContent);
        if (logger.isInfoEnabled()) {
            logger.info(requestInfo.toString());
            logger.info(requestInfo.getRequestContent());
        }
        RequestEntity requestEntity = new StringRequestEntity(requestInfo.getRequestContent(), CONTENT_TYPE, CHARSET);
        PostMethod postMethod = new PostMethod(requestInfo.getUrl());
        try {
            postMethod.setRequestEntity(requestEntity);
            this.httpClient.executeMethod(postMethod);
            String responseStr = postMethod.getResponseBodyAsString();
            return JSONObject.parseObject(responseStr);
        } catch (Exception e) {
            logger.warn("send_message_error", e);
            throw new Exception("[ Send message error ]", e);
        } finally {
            postMethod.releaseConnection();
        }
    }




    /**
     * 获取文件的临时url
     * @param fileKey 文件的file key (在上传的时候获取，或者在接收消息的时候由云客服发送过来的)
     * @return
     * success 返回内容：
     * <ul>
     *     <li>type : image/video/voice</li> 文件类型
     *     <li>mediaUrl: "https://oss.aliyun.com/example/url</li> 文件临时url
     *     <li>expiretime : 12344312</li> url过期时间
     * </ul>
     * error 返回内容：
     * <ul>
     *     <li>code : 404</li> 错误码
     *     <li>msg : "error message"</li> 错误消息提示
     * </ul>
     * @throws IOException 异常内容 一般情况下都是httpClient抛出的
     */
    public String fetchFileUrl(String fileKey) throws IOException {
        StringBuffer urlBuffer = new StringBuffer();
        long timestamp = System.currentTimeMillis();

        String digest = hmac_sha1(fileKey + timestamp, requestInfo.getKey());
        String mediaUrl = urlBuffer.append(requestInfo.getUrl()).append(CustomerServiceConfig.HOST_URL).append(CustomerServiceConfig.FETCH_FILE_API_PATH).append(
                "?fileKey=").append(fileKey).append("&tntInstId=").append(requestInfo.getTntInstId()).append("&timestamp=").append(timestamp).append("&digest=").append(digest).toString();
        if (logger.isInfoEnabled()) {
            logger.info("media url is : " + mediaUrl);
        }
        GetMethod getMethod= new GetMethod(mediaUrl);
        try {
            this.httpClient.executeMethod(getMethod);
            String responseStr = getMethod.getResponseBodyAsString();
            return responseStr;
        } catch (HttpException e) {
            throw new HttpException("[ fetch file url http error ] fileKey = " + fileKey, e);
        } catch (IOException e) {
            throw new IOException("[ fetch file url error ] fileKey = " + fileKey, e);
        } finally {
            getMethod.releaseConnection();
        }
    }


}